glm.fsreg_2 <- function(target, dataset, iniset = NULL, wei = NULL, threshold = 0.05, tol = 2, heavy, robust = FALSE, ncores = 1) {
  
  ## target can be Real valued (normal), binary (binomial) or counts (poisson)
  ## dataset is a matrix or a data.frame with the predictor variables
  ## test is the test used, but not really required because depending on the data this will be decided.
  ## there is no hrm is psecifying this though
  ## threshold is the level of significance
  ## method can be either BIC or adjrsq (for non-robust linear models only). The BIC is a consistent method for selecting
  ## models, thus we also use it to avoid overfitting
  ## stopping is based on "BIC"
  ## tol is the tolerance value for the method. If BIC is used as the stopping rule, the default is 2, but usually can be 2 or 4.
  ## If BIC is used as a way to proceed, the tol is 0.
  ## robust is for robust modelling. TRUE or FALSE
  ## ncores is for parallel processing 
  
  threshold <- log(threshold)  ## log of the significance level
  
  p <- ncol(dataset)  ## number of variables
  devi <- dof <- numeric( p )  
  moda <- list()
  k <- 1   ## counter
  n <- length(target)  ## sample size
  tool <- numeric( min(n, p) )
  con <- log(n)
  
  #########
  ## if it is binomial or poisson regression
  #########
  
  if ( is.null(iniset) ) {
    da <- 0
    pa <- 0
    
  } else {
    pa <- NCOL(iniset)
    da <- 1:pa
    dataset <- cbind(iniset, dataset)
  }  
  
  
  if ( is.matrix(target)  &  NCOL(target) == 2 )  {
    
    ci_test <- "testIndBinom"
    y <- target[, 1]
    wei <- target[, 2]
    
    runtime <- proc.time()
    
    devi = dof = numeric(p)
    #if ( robust == FALSE ) {
    if ( pa == 0 ) {
      ini <- glm( y / wei ~ 1, weights = wei, family = binomial, y = FALSE, model = FALSE )$deviance  ## residual deviance
    } else  ini <- glm(y / wei ~., data = as.data.frame( iniset ), weights = wei, family = binomial, y = FALSE, model = FALSE )$deviance  ## residual deviance
    #} else {
    #  ini = robust::glmRob( target ~ 1, family = oiko, maxit = maxit )$deviance  ## residual deviance
    #}
    
    if (ncores <= 1) {
      #if ( robust == FALSE ) {  ## Non robust
      for (i in 1:p) {
        mi <- glm( y / wei ~ . , as.data.frame( dataset[, c(da, pa + i)] ), weights = wei, family = binomial, y = FALSE, model = FALSE )
        devi[i] <- mi$deviance
        dof[i] = length( coef( mi ) ) 
      }
      
      #} else {  ## Robust
      #  for (i in 1:p) { 
      #    mi <- robust::glmRob( target ~ dataset[, i], family = oiko, maxit = maxit )
      #    devi[i] <- mi$deviance
      #    dof[i] = length( coef( mi ) )          
      #  }
      #}
      
      stat = ini - devi
      pval = pchisq( stat, dof - 1, lower.tail = FALSE, log.p = TRUE )
      
    } else {
      #if ( robust == FALSE ) {  ## Non robust
      cl <- makePSOCKcluster(ncores)
      registerDoParallel(cl)
      mata <- matrix(0, p, 2)
      mod <- foreach( i = 1:p, .combine = rbind) %dopar% {
        ww <- glm( y / wei ~., data = as.data.frame( dataset[, c(da, pa + i)] ), weights = wei, family = binomial )
        mata[i, ] <- c( ww$deviance, length( coef( ww ) )  )
      }
      
      stopCluster(cl)
      
      #       } else {  ## Robust
      #         cl <- makePSOCKcluster(ncores)
      #         registerDoParallel(cl)
      #         mata <- matrix(0, p, 2)
      #         mod <- foreach( i = 1:p, .combine = rbind, .export = "glmRob", .packages = "robust" ) %dopar% {
      #           ww <- robust::glmRob( target ~ dataset[, i], family = oiko, maxit = maxit  )
      #           mata[i, ] <- c( ww$deviance, length( coef( ww ) )  )
      #         }
      # 
      #         stopCluster(cl)
      #       }
      
      stat <- ini - mod[, 1]
      pval <- pchisq( stat, mod[, 2] - 1, lower.tail = FALSE, log.p = TRUE )
    }
    
    mat <- cbind(1:p, pval, stat) 
    
    colnames(mat)[1] <- "variables"
    rownames(mat) <- 1:p
    
    sel <- which.min(pval)
    info <- matrix( c( 1e300, 0, 0 ), ncol = 3 )
    sela <- sel
    
    if ( mat[sel, 2] < threshold ) {
      info[k, ] <- mat[sel, ]
      mat <- mat[-sel, ] 
      if ( !is.matrix(mat) )  mat <- matrix(mat, ncol = 3) 
      #if ( robust == FALSE ) {
      mi <- glm( y / wei ~., data = as.data.frame( dataset[, c(da, sel) ] ), weights = wei, family = binomial, y = FALSE, model = FALSE )
      tool[k] <- BIC( mi )
      #} else {
      #  mi <- robust::glmRob( target ~ dataset[, sel], family = oiko, maxit = maxit )
      #  tool[1] <- mi$deviance + length( coef( mi ) ) * log(n)
      #}
      moda[[ k ]] <- mi
    }
    
    
    ############
    ###       k equals 2
    ############ 
    
    
    if ( info[k, 2] < threshold  &  nrow(mat) > 0 ) {
      
      k <- k + 1
      pn <- p - k + 1   
      
      ini <- moda[[ 1 ]]$deviance  ## residual deviance
      do <- length( coef( moda[[ 1 ]]  ) ) 
      
      if ( ncores <= 1 ) {
        devi <- dof <- numeric(pn)
        #if ( robust == FALSE ) {  ## Non robust
        for ( i in 1:pn ) {
          ww <- glm( y / wei ~., data = as.data.frame( dataset[, c(da, sela, mat[pa + i, 1]) ] ), weights = wei, family = binomial, y = FALSE, model = FALSE )
          devi[i] <- ww$deviance
          dof[i] <- length( coef( ww ) )          
        }
        
        # } else {  ## Robust
        #   for ( i in 1:pn ) {
        #     ww <- robust::glmRob( target ~., data = as.data.frame( dataset[, c(sela, mat[i, 1]) ] ), family = oiko, maxit = maxit )
        #     devi[i] <- ww$deviance
        #     dof[i] = length( coef( ww ) ) 
        #   }   
        # }
        
        stat <- ini - devi
        pval <- pchisq( stat, dof - do, lower.tail = FALSE, log.p = TRUE )
        
      } else {
        
        #if ( robust == FALSE ) {  ## Non robust
        cl <- makePSOCKcluster(ncores)
        registerDoParallel(cl)
        mata <- matrix(0, pn, 2)  
        mod <- foreach( i = 1:pn, .combine = rbind) %dopar% {
          ww <- glm( y / wei ~., data = as.data.frame( dataset[, c(da, sela, mat[pa + i, 1]) ] ), weights = wei, family = binomial )
          mata[i, ] <- c( ww$deviance, length( coef( ww ) ) )
        }
        
        stopCluster(cl)
        
        # } else {  ## Robust
        #   cl <- makePSOCKcluster(ncores)
        #   registerDoParallel(cl)
        #   mata = matrix(0, pn, 2)  
        #   mod <- foreach( i = 1:pn, .combine = rbind, .export = "glmRob", .packages = "robust" ) %dopar% {
        #     ww <- robust::glmRob( target ~., data = as.data.frame( dataset[, c(sela, mat[i, 1]) ] ), family = oiko, maxit = maxit )
        #     mata[i, ] <- c( ww$deviance, length( coef( ww ) ) )
        #   }
        # 
        #   stopCluster(cl)
        # }
        
        stat <- ini - mod[, 1]
        pval <- pchisq( stat, mod[, 2] - do, lower.tail = FALSE, log.p = TRUE )
        
      }
      
      mat[, 2:3] <- cbind(pval, stat)
      
      ina <- which.min(mat[, 2])
      sel <- mat[ina, 1]    
      
      if ( mat[ina, 2] < threshold ) {
        #if ( robust == FALSE ) {
        ma <- glm( y / wei ~., data=as.data.frame( dataset[, c(da, sela, sel) ] ), weights = wei, family = binomial, y = FALSE, model = FALSE )
        tool[k] <- BIC( ma )
        #} else {
        #  ma <- robust::glmRob( target ~  dataset[, sela] + dataset[, sel], family = oiko, maxit = maxit )
        #  tool[k] <- ma$deviance + length( coef( ma ) ) * log(n)
        #}
        
        if ( tool[ k - 1 ] - tool[ k ] <= tol ) {
          info <- info
          
        } else { 
          info <- rbind(info, c( mat[ina, ] ) )
          sela <- info[, 1]
          mat <- mat[-ina , ] 
          if ( !is.matrix(mat) )   mat <- matrix(mat, ncol = 3) 
          moda[[ k ]] <- ma
        }
        
      } else   info <- info
      
    }
    
    
    ############
    ###       k greater than 2
    ############ 
    
    
    if ( nrow(info) > 1  &  nrow(mat) > 0 ) {
      while ( ( info[k, 2] < threshold ) &  ( k < n ) & ( tool[ k - 1 ] - tool[ k ] > tol ) & ( nrow(mat) > 0 ) )  {
        
        ini <- moda[[ k ]]$deviance  ## residual deviance
        do <- length( coef( moda[[ k ]]  ) ) 
        
        k <- k + 1   
        pn <- p - k  + 1
        
        if (ncores <= 1) {  
          devi = dof = numeric(pn) 
          #if ( robust == FALSE ) {  ## Non robust
          for ( i in 1:pn ) {
            ma <- glm( y / wei ~., data = as.data.frame( dataset[, c(da, sela, mat[pa + i, 1] ) ] ), weights = wei, family = binomial, y = FALSE, model = FALSE )
            devi[i] <- ma$deviance
            dof[i] <- length( coef( ma ) ) 
          }
          
          # } else {  ## Robust
          #   for ( i in 1:pn ) {
          #     ma <- robust::glmRob( target ~., data = as.data.frame( dataset[, c(sela, mat[i, 1]) ] ), family = oiko, maxit = maxit )
          #     devi[i] <- ma$deviance
          #     dof[i] = length( coef( ma ) ) 
          #   }
          # }
          
          stat <- ini - devi
          pval <- pchisq( stat, dof - do, lower.tail = FALSE, log.p = TRUE )
          
        } else {
          #if ( robust == FALSE ) {  ## Non robust
          cl <- makePSOCKcluster(ncores)
          registerDoParallel(cl)
          devi <- dof <- numeric(pn)
          mata <- matrix(0, pn, 2)  
          mod <- foreach( i = 1:pn, .combine = rbind) %dopar% {
            ww <- glm( y / wei ~., data = as.data.frame( dataset[, c(da, sela, mat[pa + i, 1]) ] ), weights = wei, family = binomial, y = FALSE, model = FALSE )
            mata[i, ] <- c( ww$deviance, length( coef( ww ) ) )
          }
          
          stopCluster(cl)
          
          # } else {  ## Robust
          #   cl <- makePSOCKcluster(ncores)
          #   registerDoParallel(cl)
          #   mata = matrix(0, pn, 2)  
          #   mod <- foreach( i = 1:pn, .combine = rbind, .export = "glmRob", .packages = "robust" ) %dopar% {
          #     ww <- robust::glmRob( target ~., data = as.data.frame( dataset[, c(sela, mat[i, 1]) ] ), family = oiko, maxit = maxit )
          #     mata[i, ] <- c( ww$deviance, length( coef( ww ) ) )
          #   }
          # 
          #   stopCluster(cl)
          # }
          
          stat <- ini - mod[, 1]
          pval <- pchisq( stat, mod[, 2] - do, lower.tail = FALSE, log.p = TRUE )
          
        }
        
        mat[, 2:3] <- cbind(pval, stat)
        
        ina <- which.min(mat[, 2])
        sel <- mat[ina, 1]    
        
        if ( mat[ina, 2] < threshold ) {
          #if ( robust == FALSE ) {
          ma <- glm( y / wei ~., data = as.data.frame( dataset[, c(da, sela, sel) ] ), weights = wei, family = binomial, y = FALSE, model = FALSE )
          tool[k] <- BIC( ma )
          #} else {
          #  ma <- robust::glmRob( target ~., data = as.data.frame( dataset[, c(sela, sel) ] ), family = oiko, maxit = maxit )
          #  tool[k] <- ma$deviance + length( coef( ma ) ) * log(n)
          #} 
          if ( tool[ k - 1 ] - tool[ k  ] < tol ) {
            info <- rbind(info, c( 1e300, 0, 0 ) )
            
          } else { 
            info <- rbind( info, mat[ina, ] )
            sela <- info[, 1]
            mat <- mat[-ina , ]
            if ( !is.matrix(mat) )  mat <- matrix(mat, ncol = 3) 
            moda[[ k ]] <- ma
          } 
          
        } else   info <- rbind(info, c( 1e300, 0, 0 ) )
        
      }
      
    } 
    
    runtime <- proc.time() - runtime
    
    d <- length(moda)
    final <- NULL
    models <- NULL
    
    if ( d == 0 ) {
      final <- glm( y / wei ~., data = as.data.frame( iniset ), weights = wei, family = binomial, y = FALSE, model = FALSE )
      info <- NULL
      
    } else {
      
      if ( d >= 1 ) {
        models <- NULL
        xx <- as.data.frame( dataset[, c(da, sela) ] )
        if ( pa == 0 ) {
          colnames(xx) <- paste("V", sela, sep = "") 
        } else  colnames(xx) <- c( paste("X", da, sep = ""),  paste("V", sela, sep = "") )
        
        if ( d == 1 ) {
          
          xx <- as.data.frame( dataset[, c(da, sela)] )
          if ( pa == 0 ) {
            colnames(xx) <- paste("V", sela, sep = "") 
          } else  colnames(xx) <- c( paste("X", da, sep = ""),  paste("V", sela, sep = "") )
          
          #if ( robust == FALSE ) {
          models[[ 1 ]] <- final <- glm( y / wei ~., data = as.data.frame( xx ), weights = wei, family = binomial, y = FALSE, model = FALSE )
          #} else {
          #  models[[ 1]] <- final <- robust::glmRob( target ~., data = as.data.frame( xx ), family = oiko, maxit = maxit )
          #}
          
        } else {
          for ( i in 1:c(pa + d) ) {
            #if ( robust == FALSE ) {
            models[[ i ]] <- glm( y / wei ~., data = as.data.frame( xx[, 1:i] ), weights = wei, family = binomial, y = FALSE, model = FALSE )
            #} else { 
            #  models[[ i ]] <- robust::glmRob( target ~., data = as.data.frame( xx[, 1:i]), family = oiko, maxit = maxit )
            #}
          }
        }
        
        final <- summary( models[[ d ]] )
        
        info <- info[1:d, ]
        if ( d == 1 )  info <- matrix(info, nrow = 1)
        info <- cbind( info, tool[ 1:d ] ) 
        colnames(info) <- c( "variables", "log.p-value", "stat", "BIC" )
        rownames(info) <- info[, 1]
      }
      
    }    
    
    result <- list(mat = t(mat), info = info, models = models, final = final, runtime = runtime ) 
   
    
    #############################
    #############################
    
  } else {
    

	
    ####################
    ### Logistic or poisson regression
    ####################
    
    
  if ( length( Rfast::sort_unique(target) ) == 2 ) {
    oiko <- binomial(logit)  ## binomial regression
    ci_test <- "testIndLogistic"
  } else  {
    ci_test <- "testIndLogistic"
    oiko <- poisson(log)  ## poisson regression
  }
    
  runtime <- proc.time()
  
  devi = dof = numeric(p)
  
  #if ( robust == FALSE ) {
  if ( pa == 0 ) {
  
    if ( !heavy ) {
      ini <- glm( target ~ 1, family = oiko, weights = wei, y = FALSE, model = FALSE )$deviance  ## residual deviance
	  } else {
	    ci_test <- "testIndSpeedglm"
	    ini <- speedglm::speedglm( target ~ 1, data = data.frame(dataset), family = oiko, weights = wei )$deviance  ## residual deviance
	  }   
	
  } else {
    if ( !heavy ) {
      ini <- glm(target ~., data = data.frame( iniset ), family = oiko, y = FALSE, model = FALSE )$deviance  ## residual deviance
	  } else {
	    ci_test <- "testIndSpeedglm"
	    ini <- speedglm::speedglm(target ~., data = data.frame( iniset ), family = oiko, weights = wei )$deviance  ## residual deviance
	  }
	
  }
  
  #} else {
  #  ini = robust::glmRob( target ~ 1, family = oiko, maxit = maxit )$deviance  ## residual deviance
  #}
  
  if (ncores <= 1) {
    #if ( robust == FALSE ) {  ## Non robust
	  if ( !heavy  ) {
      for (i in 1:p) {
        mi <- glm( target ~ ., data.frame( dataset[, c(da, pa + i)] ), family = oiko, weights= wei, y = FALSE, model = FALSE )
        devi[i] <- mi$deviance
        dof[i] <- length( coef( mi ) ) 
      }
    
	} else {
	  for (i in 1:p) {
        mi <- speedglm::speedglm( target ~ ., data.frame( dataset[, c(da, pa + i)] ), family = oiko, weights = wei )
        devi[i] <- mi$deviance
        dof[i] <- length( coef( mi ) ) 
      }
	  
	}
    #} else {  ## Robust
    #  for (i in 1:p) { 
    #    mi <- robust::glmRob( target ~ dataset[, i], family = oiko, maxit = maxit )
    #    devi[i] <- mi$deviance
    #    dof[i] = length( coef( mi ) )          
    #  }
    #}
    
    stat <- ini - devi
    pval <- pchisq( stat, dof - 1, lower.tail = FALSE, log.p = TRUE )
    
  } else {
    #if ( robust == FALSE ) {  ## Non robust
	if ( !heavy ) { 
      cl <- makePSOCKcluster(ncores)
      registerDoParallel(cl)
      mata <- matrix(0, p, 2)
      mod <- foreach( i = 1:p, .combine = rbind) %dopar% {
        ww <- glm( target ~., data = data.frame( dataset[, c(da, pa + i)] ), family = oiko, weights = wei, y = FALSE, model = FALSE )
        mata[i, ] <- c( ww$deviance, length( coef( ww ) )  )
      }
    
      stopCluster(cl)
    
	} else {
      cl <- makePSOCKcluster(ncores)
      registerDoParallel(cl)
      mata <- matrix(0, p, 2)
      mod <- foreach( i = 1:p, .combine = rbind, .export = "speedglm", .packages = "speedglm") %dopar% {
        ww <- speedglm::speedglm( target ~., data = data.frame( dataset[, c(da, pa + i)] ), family = oiko, weights = wei )
        mata[i, ] <- c( ww$deviance, length( coef( ww ) )  )
      }
    
      stopCluster(cl)	
	}
	
    #       } else {  ## Robust
    #         cl <- makePSOCKcluster(ncores)
    #         registerDoParallel(cl)
    #         mata <- matrix(0, p, 2)
    #         mod <- foreach( i = 1:p, .combine = rbind, .export = "glmRob", .packages = "robust" ) %dopar% {
    #           ww <- robust::glmRob( target ~ dataset[, i], family = oiko, maxit = maxit  )
    #           mata[i, ] <- c( ww$deviance, length( coef( ww ) )  )
    #         }
    # 
    #         stopCluster(cl)
    #       }
    
    stat <- ini - mod[, 1]
    pval <- pchisq( stat, mod[, 2] - 1, lower.tail = FALSE, log.p = TRUE )
  }
  
  mat <- cbind(1:p, pval, stat) 
  
  colnames(mat)[1] <- "variables"
  rownames(mat) <- 1:p
  
  sel <- which.min(pval)
  info <- matrix( c( 1e300, 0, 0 ), ncol = 3 )
  sela <- sel
  
  if ( mat[sel, 2] < threshold ) {
    info[k, ] <- mat[sel, ]
    mat <- mat[-sel, ] 
    if ( !is.matrix(mat) )   mat <- matrix(mat, ncol = 3) 

    #if ( robust == FALSE ) {
	   if ( !heavy ) {
      mi <- glm( target ~., data = as.data.frame( dataset[, c(da, sel) ] ), family = oiko, weights = wei, y = FALSE, model = FALSE )
      tool[k] <- BIC( mi )
	  } else {
	    mi <- speedglm::speedglm( target ~., data = as.data.frame( dataset[, c(da, sel) ] ), family = oiko, weights = wei )
      tool[k] <-  - 2 * logLik(mi) + length( coef(mi) ) * con
  	}
    #} else {
    #  mi <- robust::glmRob( target ~ dataset[, sel], family = oiko, maxit = maxit )
    #  tool[1] <- mi$deviance + length( coef( mi ) ) * log(n)
    #}
    moda[[ k ]] <- mi
  }
  
  
  ############
  ###       k equals 2
  ############ 
  
  
  if ( info[k, 2] < threshold  &  nrow(mat) > 0 ) {
    
    k <- k + 1
    pn <- p - k + 1   
    
    ini <- mi$deviance  ## residual deviance
    do <- length( coef( mi ) ) 
    
    if ( ncores <= 1 ) {
      devi <- dof <- numeric(pn)
      #if ( robust == FALSE ) {  ## Non robust
	  if ( !heavy ) {
        for ( i in 1:pn ) {
          ww <- glm( target ~., data = as.data.frame( dataset[, c(da, sela, mat[pa + i, 1]) ] ), family = oiko, weights = wei, y = FALSE, model = FALSE )
          devi[i] <- ww$deviance
          dof[i] <- length( coef( ww ) )          
        }
		
	  } else {
        for ( i in 1:pn ) {
          ww <- speedglm::speedglm( target ~., data = as.data.frame( dataset[, c(da, sela, mat[pa + i, 1]) ] ), family = oiko, weights = wei )
          devi[i] <- ww$deviance
          dof[i] <- length( coef( ww ) )          
        }	  
	  }	
      
      # } else {  ## Robust
      #   for ( i in 1:pn ) {
      #     ww <- robust::glmRob( target ~., data = as.data.frame( dataset[, c(sela, mat[i, 1]) ] ), family = oiko, maxit = maxit )
      #     devi[i] <- ww$deviance
      #     dof[i] = length( coef( ww ) ) 
      #   }   
      # }
      
      stat <- ini - devi
      pval <- pchisq( stat, dof - do, lower.tail = FALSE, log.p = TRUE )
      
    } else {
      
      #if ( robust == FALSE ) {  ## Non robust
	  if ( !heavy ) {
        cl <- makePSOCKcluster(ncores)
        registerDoParallel(cl)
        mata <- matrix(0, pn, 2)  
        mod <- foreach( i = 1:pn, .combine = rbind) %dopar% {
          ww <- glm( target ~., data = as.data.frame( dataset[, c(da, sela, mat[pa + i, 1]) ] ), family = oiko, weights = wei, y = FALSE, model = FALSE )
          mata[i, ] <- c( ww$deviance, length( coef( ww ) ) )
        }
      
        stopCluster(cl)
	  
      } else {
        cl <- makePSOCKcluster(ncores)
        registerDoParallel(cl)
        mata <- matrix(0, pn, 2)  
        mod <- foreach( i = 1:pn, .combine = rbind, .export = "speedglm", .packages = "speedglm") %dopar% {
          ww <- speedglm::speedglm( target ~., data = as.data.frame( dataset[, c(da, sela, mat[pa + i, 1]) ] ), family = oiko, weights = wei )
          mata[i, ] <- c( ww$deviance, length( coef( ww ) ) )
        }
      
        stopCluster(cl)	  
	  }
	  
	  
      # } else {  ## Robust
      #   cl <- makePSOCKcluster(ncores)
      #   registerDoParallel(cl)
      #   mata = matrix(0, pn, 2)  
      #   mod <- foreach( i = 1:pn, .combine = rbind, .export = "glmRob", .packages = "robust" ) %dopar% {
      #     ww <- robust::glmRob( target ~., data = as.data.frame( dataset[, c(sela, mat[i, 1]) ] ), family = oiko, maxit = maxit )
      #     mata[i, ] <- c( ww$deviance, length( coef( ww ) ) )
      #   }
      # 
      #   stopCluster(cl)
      # }
      
      stat <- ini - mod[, 1]
      pval <- pchisq( stat, mod[, 2] - do, lower.tail = FALSE, log.p = TRUE )
      
    }
    
    mat[, 2:3] <- cbind(pval, stat)
    
    ina <- which.min(mat[, 2])
    sel <- mat[ina, 1]    
    
    if ( mat[ina, 2] < threshold ) {
      #if ( robust == FALSE ) {
	    if ( !heavy )  {
        ma <- glm( target ~., data=as.data.frame( dataset[, c(da, sela, sel) ] ), family = oiko, weights = wei, y = FALSE, model = FALSE )
        tool[k] <- BIC( ma )
	    } else {
        ma <- speedglm::speedglm( target ~., data=as.data.frame( dataset[, c(da, sela, sel) ] ), family = oiko, weights = wei )
        tool[k] <-  - 2 * logLik(ma) + length( coef(ma) ) * con	  
	    }	
      #} else {
      #  ma <- robust::glmRob( target ~  dataset[, sela] + dataset[, sel], family = oiko, maxit = maxit )
      #  tool[k] <- ma$deviance + length( coef( ma ) ) * log(n)
      #}
      
      if ( tool[ k - 1 ] - tool[ k ] <= tol ) {
        info <- rbind(info, c( 1e300, 0, 0 ) )
        
      } else { 
        info <- rbind(info, c( mat[ina, ] ) )
        sela <- info[, 1]
        mat <- mat[-ina , ] 
        if ( !is.matrix(mat) )   mat <- matrix(mat, ncol = 3) 
        moda[[ k ]] <- ma
      }
      
    } else    info <- info
    
  }
  
  
  ############
  ###       k greater than 2
  ############ 
  
  
  if ( nrow(info) > 1  &  nrow(mat) > 0 ) {
    while ( ( info[k, 2] < threshold ) &  ( k < n ) & ( tool[ k - 1 ] - tool[ k ] > tol ) & ( nrow(mat) > 0 ) )  {
      
      ini <- moda[[ k ]]$deviance  ## residual deviance
      do <- length( coef( moda[[ k ]]  ) ) 
      
      k <- k + 1   
      pn <- p - k  + 1
      
      if (ncores <= 1) {  
        devi <- dof <- numeric(pn) 
        #if ( robust == FALSE ) {  ## Non robust
		    if ( !heavy ) {
          for ( i in 1:pn ) {
            ma <- glm( target ~., data = as.data.frame( dataset[, c(da, sela, mat[pa + i, 1] ) ] ), family = oiko, weights = wei, y = FALSE, model = FALSE )
            devi[i] <- ma$deviance
            dof[i] <- length( coef( ma ) ) 
          }
		
        } else {
          for ( i in 1:pn ) {
            ma <- speedglm:speedglm( target ~., data = as.data.frame( dataset[, c(da, sela, mat[pa + i, 1] ) ] ), family = oiko, weights = wei )
            devi[i] <- ma$deviance
            dof[i] <- length( coef( ma ) ) 
          }		
		    }
        # } else {  ## Robust
        #   for ( i in 1:pn ) {
        #     ma <- robust::glmRob( target ~., data = as.data.frame( dataset[, c(sela, mat[i, 1]) ] ), family = oiko, maxit = maxit )
        #     devi[i] <- ma$deviance
        #     dof[i] = length( coef( ma ) ) 
        #   }
        # }
        
        stat <- ini - devi
        pval <- pchisq( stat, dof - do, lower.tail = FALSE, log.p = TRUE )
        
      } else {
        #if ( robust == FALSE ) {  ## Non robust
		    if ( !heavy )  {
          cl <- makePSOCKcluster(ncores)
          registerDoParallel(cl)
          devi <- dof <- numeric(pn)
          mata <- matrix(0, pn, 2)  
          mod <- foreach( i = 1:pn, .combine = rbind) %dopar% {
            ww <- glm( target ~., data = as.data.frame( dataset[, c(da, sela, mat[pa + i, 1]) ] ), family = oiko, weights = wei, y = FALSE, model = FALSE )
            mata[i, ] <- c( ww$deviance, length( coef( ww ) ) )
          }
        
          stopCluster(cl)
        } else {
          cl <- makePSOCKcluster(ncores)
          registerDoParallel(cl)
          devi <- dof <- numeric(pn)
          mata <- matrix(0, pn, 2)  
          mod <- foreach( i = 1:pn, .combine = rbind ,.export = "speedglm", .packages = "speedglm") %dopar% {
            ww <- speedglm( target ~., data = as.data.frame( dataset[, c(da, sela, mat[pa + i, 1]) ] ), family = oiko, weights = wei )
            mata[i, ] <- c( ww$deviance, length( coef( ww ) ) )
          }
        
          stopCluster(cl)		
	    	}
        
        # } else {  ## Robust
        #   cl <- makePSOCKcluster(ncores)
        #   registerDoParallel(cl)
        #   mata = matrix(0, pn, 2)  
        #   mod <- foreach( i = 1:pn, .combine = rbind, .export = "glmRob", .packages = "robust" ) %dopar% {
        #     ww <- robust::glmRob( target ~., data = as.data.frame( dataset[, c(sela, mat[i, 1]) ] ), family = oiko, maxit = maxit )
        #     mata[i, ] <- c( ww$deviance, length( coef( ww ) ) )
        #   }
        # 
        #   stopCluster(cl)
        # }
        
        stat <- ini - mod[, 1]
        pval <- pchisq( stat, mod[, 2] - do, lower.tail = FALSE, log.p = TRUE )
        
      }
      
      mat[, 2:3] <- cbind(pval, stat)
      
      ina <- which.min(mat[, 2])
      sel <- mat[ina, 1]    
      
      if ( mat[ina, 2] < threshold ) {
        #if ( robust == FALSE ) {
		    if ( !heavy ) {
          ma <- glm( target ~., data = as.data.frame( dataset[, c(da, sela, sel) ] ), family = oiko, weights = wei, y = FALSE, model = FALSE )
          tool[k] <- BIC( ma )
        } else {
          ma <- speedglm::speedglm( target ~., data = as.data.frame( dataset[, c(da, sela, sel) ] ), family = oiko, weights = wei )
          tool[k] <-  - 2 * logLik(ma) + length( coef(ma) ) * con		
		    }
		#} else {
        #  ma <- robust::glmRob( target ~., data = as.data.frame( dataset[, c(sela, sel) ] ), family = oiko, maxit = maxit )
        #  tool[k] <- ma$deviance + length( coef( ma ) ) * log(n)
        #} 
        if ( tool[ k - 1 ] - tool[ k  ] < tol ) {
          info <- rbind(info, c( 1e300, 0, 0 ) )
          
        } else { 
          info <- rbind( info, mat[ina, ] )
          sela <- info[, 1]
          mat <- mat[-ina , ]
          if ( !is.matrix(mat) )   mat <- matrix(mat, ncol = 3) 
          moda[[ k ]] <- ma
        } 
        
      } else   info <- rbind(info, c( 1e300, 0, 0 ) )
      
    }
    
  } 
  
  runtime <- proc.time() - runtime
  
  d <- length(moda)
  final <- NULL
  models <- NULL
  
  if ( d == 0 ) {
    if ( !heavy ) {
      final <- glm( target ~., data = as.data.frame( iniset ), family = oiko, weights = wei, y = FALSE, model = FALSE ) 
	  } else  final <- speedglm::speedglm( target ~., data = as.data.frame( iniset ), family = oiko, weights = wei ) 
    info <- NULL
    
  } else {
  
  if ( d >= 1 ) {
    models <- NULL
    xx <- as.data.frame( dataset[, c(da, sela) ] )
    if ( pa == 0 ) {
      colnames(xx) <- paste("V", sela, sep = "") 
    } else   colnames(xx) <- c( paste("X", da, sep = ""),  paste("V", sela, sep = "") )
    
    if ( d == 1 ) {
      
      xx <- as.data.frame( dataset[, c(da, sela)] )
      if ( pa == 0 ) {
        colnames(xx) <- paste("V", sela, sep = "") 
      } else  colnames(xx) <- c( paste("X", da, sep = ""),  paste("V", sela, sep = "") )
      
      #if ( robust == FALSE ) {
	    if ( !heavy ) {
        models[[ 1 ]] <- final <- glm( target ~., data = as.data.frame( xx ), family = oiko, weights = wei, y = FALSE, model = FALSE )
	    } else  models[[ 1 ]] <- final <- speedglm::speedglm( target ~., data = as.data.frame( xx ), family = oiko, weights = wei )
	 	
      #} else {
      #  models[[ 1]] <- final <- robust::glmRob( target ~., data = as.data.frame( xx ), family = oiko, maxit = maxit )
      #}
      
    } else {
      for ( i in 1:c(pa + d) ) {
        #if ( robust == FALSE ) {
		if ( !heavy ) {
        models[[ i ]] <- glm( target ~., data = as.data.frame( xx[, 1:i] ), family = oiko, weights = wei, y = FALSE, model = FALSE )
		} else  models[[ i ]] <- speedglm::speedglm( target ~., data = as.data.frame( xx[, 1:i] ), family = oiko, weights = wei )
        #} else { 
        #  models[[ i ]] <- robust::glmRob( target ~., data = as.data.frame( xx[, 1:i]), family = oiko, maxit = maxit )
        #}
      }
    }
    
    final <- summary( models[[ d ]] )
    
    info <- info[1:d, ]
    if ( d == 1 )  info <- matrix(info, nrow = 1)
    info <- cbind( info, tool[ 1:d ] ) 
    colnames(info) <- c( "variables", "log.p-value", "stat", "BIC" )
    rownames(info) <- info[, 1]
  }
  
  }    
  
  result <-   list(mat = t(mat), info = info, models = models, final = final, ci_test = ci_test, runtime = runtime ) 

  
  }

  
  result  
  
}    










